from .knapsack import Knapsack, SlackFreeKnapsack
from .maximumcut import MaximumCut
from .minimumvertexcover import MinimumVertexCover
from .numberpartition import NumberPartition
from .shortestpath import ShortestPath
from .tsp import TSP, TSP_LP
from .portfoliooptimization import PortfolioOptimization
from .maximalindependentset import MIS
from .binpacking import BinPacking
from .vehiclerouting import VRP
from .sherrington_kirkpatrick import SK
from .kcolor import KColor
from .converters import FromDocplex2IsingModel
from .qubo import QUBO
